from typing import Dict, Iterable, List, TextIO

from dataclasses import dataclass
from io import TextIOWrapper
from pathlib import Path

import numpy as np
import orjson as json

from multipledispatch import dispatch
from numpy import ndarray as array

from .RunData import RunData
from .TimeStep import TimeStep


@dataclass
class RunDataCorrected:
    timestamp: str
    run_name: str
    flux_name: str
    dose_rate_type: str
    dose_rate_distance: float


@dataclass
class Inventory:
    run_data: RunData
    inventory_data: List[TimeStep]

    @classmethod
    def from_json(cls, json_dict: Dict):
        json_run_data = json_dict.pop("run_data")
        run_data = RunData.from_json(json_run_data)
        json_inventory_data = json_dict.pop("inventory_data")

        def create_json_inventory_data_mapper():
            prev_irradiation_time = prev_cooling_time = prev_elapsed_time = 0.0
            number = 1

            def json_inventory_data_mapper(jts: Dict):
                nonlocal number, prev_irradiation_time, prev_cooling_time, prev_elapsed_time
                ts = TimeStep.from_json(jts)
                duration = ts.irradiation_time - prev_irradiation_time
                if 0.0 == duration:
                    duration = ts.cooling_time - prev_cooling_time
                assert (
                    0.0 <= duration
                ), "Irradiation and cooling times in FISPACT JSON output should be monotonically increasing"
                ts.duration = duration
                prev_elapsed_time = ts.elapsed_time = prev_elapsed_time + duration
                if duration == 0.0:
                    ts.flux = 0.0
                ts.number = number
                number += 1
                prev_irradiation_time, prev_cooling_time = (
                    ts.irradiation_time,
                    ts.cooling_time,
                )
                return ts

            return json_inventory_data_mapper

        inventory_data = list(
            map(create_json_inventory_data_mapper(), json_inventory_data)
        )

        return cls(run_data, inventory_data)

    @property
    def meta_info(self) -> RunDataCorrected:
        ts = self.inventory_data[-1]  # The dose_rate in the first timestep is empty.
        dose_rate_type, dose_rate_distance = ts.dose_rate_type_and_distance
        return RunDataCorrected(
            self.run_data.timestamp,
            self.run_data.run_name,
            self.run_data.flux_name,
            dose_rate_type,
            dose_rate_distance,
        )

    # def as_array(self, time_step_extractor=total_extractor) -> array:
    #     return np.vstack(list(map(time_step_extractor, self.inventory_data)))

    def __iter__(self) -> Iterable[TimeStep]:
        return iter(self.inventory_data)

    def __len__(self) -> int:
        return len(self.inventory_data)

    def __getitem__(self, item):
        return self.inventory_data[item]


@dispatch(list)
def extract_times(time_steps: List[TimeStep]) -> array:
    return np.fromiter(map(lambda x: x.elapsed_time, time_steps), dtype=float)


@dispatch(Inventory)
def extract_times(inventory: Inventory) -> array:
    return extract_times(inventory.inventory_data)


@dispatch(str)
def from_json(text: str):
    json_dict = json.loads(text)
    return Inventory.from_json(json_dict)


@dispatch(TextIOWrapper)
def from_json(stream: TextIO):
    return from_json(stream.read())


@dispatch(Path)
def from_json(path: Path):
    return from_json(path.read_text())



"""Classes to read a FISPACT time step attributes from JSON."""

from typing import Dict, List, Tuple

from dataclasses import dataclass, field

import numpy as np

from numpy import ndarray as array

from .Nuclide import Nuclide


@dataclass
class DoseRate:
    """Dose rate attributes."""

    type: str = ""
    distance: float = 0.0
    mass: float = 0.0
    dose: float = 0.0

    def __post_init__(self) -> None:
        """Correct wrong value coming from FISPACT"""
        # TODO dvp: check behaviour with FISPACT v.5 and try to correct scenarios.
        if self.mass == 0.0:
            # According to FISPACT manual (v.4)
            # should be 1 gram always, but FISPACT shows 0 at the first step
            # and less, than 1 in the following steps. Fixing here.
            self.mass = 1.0e-3


@dataclass
class GammaSpectrum:
    """Data on gamma emission."""

    boundaries: array
    """Energy boundaries, MeV"""
    values: array
    """Gamma emission intensity."""  # TODO dvp: identify units.

    @classmethod
    def from_json(cls, data: Dict) -> "GammaSpectrum":
        return cls(
            boundaries=np.array(data["boundaries"], dtype=float),
            values=np.array(data["values"], dtype=float),
        )


@dataclass
class TimeStep:
    """Time step attributes.

    All names must be the same as in FISPACT JSON file.
    """

    number: int = -1
    irradiation_time: float = 0.0
    cooling_time: float = 0.0
    duration: float = 0.0
    elapsed_time: float = 0.0
    flux: float = 0.0
    # TODO dvp: compute the values in case of FISPACT v.4 to maintain same logic.
    total_atoms: float = 0.0  # FISPACT 5.0
    total_activity: float = 0.0  # -/-
    alpha_activity: float = 0.0  # -/-
    beta_activity: float = 0.0  # -/-
    gamma_activity: float = 0.0  # -/-
    total_mass: float = 0.0  # -/-
    total_heat: float = 0.0
    alpha_heat: float = 0.0
    beta_heat: float = 0.0
    gamma_heat: float = 0.0
    ingestion_dose: float = 0.0
    inhalation_dose: float = 0.0
    dose_rate: DoseRate = field(default_factory=DoseRate)
    nuclides: List[Nuclide] = field(default_factory=list)
    gamma_spectrum: GammaSpectrum = None

    def __post_init__(self) -> None:
        if 0.0 == self.total_mass:  # workaround for FISPACT v.4
            # Mass of all nuclides in the TimeStep (kg)
            # TODO dvp: check for FISPACT v.5
            self.total_mass = 1e-3 * sum(map(lambda x: x.grams, self.nuclides))

    @property
    def nuclides_mass(self) -> float:
        """Synonym for total mass."""
        return self.total_mass

    @property
    def is_cooling(self) -> bool:
        """Is the time step for cooling?"""
        return self.flux == 0.0

    @property
    def dose_rate_type_and_distance(self) -> Tuple[str, float]:
        """Get Dose rate type and distance.

        These two properties are duplicated over all the TimeSteps (lack of normalization?).
        The attributes are moved to RunData on postprocessing.
        """
        return self.dose_rate.type, self.dose_rate.distance

    # def key(self):
    #     return self.number
    #
    # def properties(self) -> array:
    #     return np.array(
    #         [
    #             self.irradiation_time,
    #             self.cooling_time,
    #             self.duration,
    #             self.elapsed_time,
    #             self.flux,
    #             self.total_atoms,
    #             self.total_activity,
    #             self.alpha_activity,
    #             self.beta_activity,
    #             self.gamma_activity,
    #             self.total_mass,
    #             self.total_heat,
    #             self.alpha_heat,
    #             self.beta_heat,
    #             self.gamma_heat,
    #             self.ingestion_dose,
    #             self.inhalation_dose,
    #             # TODO dvp: don't see use case for the "dose_rate.mass" parameter, check if it's needed at all.
    #             self.dose_rate.dose,
    #         ],
    #         dtype=float,
    #     )

    @classmethod
    def from_json(cls, json_dict: Dict) -> "TimeStep":
        json_dose_rate = json_dict.pop("dose_rate")
        dose_rate = DoseRate(**json_dose_rate)
        json_nuclides = json_dict.pop("nuclides", None)
        if json_nuclides:
            nuclides = list(map(Nuclide.from_json, json_nuclides))
        else:
            nuclides = []
        json_gamma_spectrum = json_dict.pop("gamma_spectrum", None)
        if json_gamma_spectrum:
            gamma_spectrum = GammaSpectrum.from_json(json_gamma_spectrum)
        else:
            gamma_spectrum = None
        return cls(
            dose_rate=dose_rate,
            nuclides=nuclides,
            gamma_spectrum=gamma_spectrum,
            **json_dict,
        )


# TIME_STEP_PROPERTIES = list(map(lambda x: x.name, dataclasses.fields(TimeStep)))[
#     1:-3
# ] + [
#     "dose"
# ]  # omitted number and original dose_rate and nuclides, added extracted dose values
# assert TIME_STEP_PROPERTIES[0] == "irradiation_time"
# assert TIME_STEP_PROPERTIES[-2] == "inhalation_dose"
# assert len(TIME_STEP_PROPERTIES) == 18



"""Output utilities."""

from typing import Any, Iterable, TextIO

import sys


def print_cols(
    seq: Iterable[Any], fid: TextIO = sys.stdout, max_columns: int = 6, fmt: str = "{}"
) -> int:
    """Print sequence in columns.

    Args:
        seq: sequence to print
        fid: output
        max_columns: max columns in a line
        fmt: format string

    Returns:
        int: the number of the last column printed on the last row

    """
    column = 0
    for s in seq:
        print(fmt.format(s), file=fid, end="")
        column += 1
        if column == max_columns:
            print(file=fid)
            column = 0
        else:
            print(" ", file=fid, end="")
    return column



"""Utility methods to access a package data."""

from typing import Callable, cast

import inspect

from pathlib import Path

import pkg_resources as pkg


def filename_resolver(package: str = None) -> Callable[[str], str]:
    """Create method to find data file name.

    Uses resource manager to handle all the cases of the deployment.

    Args:
        package: the package below which the data is stored.
                 Optional, if not specified, the package of caller will be used.

    Returns:
        callable which appends the argument to the package folder.
    """
    package = _resolve_package(package)
    resource_manager = pkg.ResourceManager()  # type: ignore

    def func(resource: str) -> str:
        return cast(str, resource_manager.resource_filename(package, resource))

    func.__doc__ = f"Computes file names for resources located in {package}"

    return func


def path_resolver(package: str = None) -> Callable[[str], Path]:
    """Create method to find data path.

    Uses :meth:`file_resolver`.

    Args:
        package: the package below which the data is stored.
                 Optional, if not specified, the package of caller will be used.

    Returns:
        callable which appends the argument to the package folder adt returns as Path.
    """
    # Note: we should define package here to have proper offset in callers stack.
    package = _resolve_package(package)
    resolver = filename_resolver(package)

    def func(resource: str) -> Path:
        filename = resolver(resource)
        return Path(filename)

    func.__doc__ = f"Computes Path for resources located in {package}"

    return func


def _resolve_package(package: str = None) -> str:
    if package is None:
        module = inspect.getmodule(inspect.stack()[2][0])
        if module is None:
            raise ValueError("Cannot define package.")  # pragma: no cover
        package = module.__name__
    return package



"""Methods to convert Inventory to xarray DataSet.

    The dimensions  ot the nodes are:
        time_step_number
        nuclide (symbol, atomic_number, state)
        gamma_boundaries if available

    The dataset facilitates FISPACT output slicing and aggregation.

    Also, the module provides methods to scale activation properties in
    activation and time_steps nodes are provided, both for mass scaling
    (all the values proportional to mass are multiplied by scaling factor)
    and for flux scaling:
    the difference with initial state is scaled with a given factor.
"""

from typing import TextIO, Tuple, Union

from functools import reduce
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr

from mckit_nuclides.nuclides import SYMBOL_2_ATOMIC_NUMBER, get_nuclide_mass

from .Inventory import Inventory
from .Inventory import from_json as inventory_from_json
from .TimeStep import TimeStep

SCALABLE_COLUMNS = [
    "total_mass",
    "total_heat",
    "total_alpha_heat",
    "total_beta_heat",
    "total_gamma_heat",
    "total_activity",
    "total_alpha_activity",
    "total_beta_activity",
    "total_gamma_activity",
    "nuclide_atoms",
    "nuclide_grams",
    "nuclide_activity",
    "nuclide_alpha_activity",
    "nuclide_beta_activity",
    "nuclide_gamma_activity",
    "nuclide_heat",
    "nuclide_alpha_heat",
    "nuclide_beta_heat",
    "nuclide_gamma_heat",
]


def create_dataset(inventory: Inventory) -> xr.Dataset:
    def _add_time_step_record(_ds: xr.Dataset, x: TimeStep) -> xr.Dataset:
        def _make_var(value, dtype=float) -> Tuple[str, np.ndarray]:
            value = np.array([value], dtype=dtype)
            return "time_step_number", value

        def _make_nuclide_var(getter, nuclides, dtype=float) -> Tuple[str, np.ndarray]:
            return "nuclide", np.fromiter(map(getter, nuclides), dtype=dtype)

        nuclide_coordinate = pd.MultiIndex.from_tuples(
            map(lambda n: (n.element, n.isotope, n.state), x.nuclides),
            names=["element", "mass_number", "state"],
        )

        def _make_time_step_and_nuclide_var(
            getter, nuclides, dtype=float
        ) -> Tuple[Tuple[str, str], np.ndarray]:
            _data = np.fromiter(map(getter, nuclides), dtype=dtype)
            return ("time_step_number", "nuclide"), _data.reshape(1, _data.size)

        data_vars = dict(
            irradiation_time=_make_var(x.irradiation_time),
            cooling_time=_make_var(x.cooling_time),
            duration=_make_var(x.duration),
            flux=_make_var(x.flux),
            total_atoms=_make_var(x.total_atoms),
            total_activity=_make_var(x.total_activity),
            total_alpha_activity=_make_var(x.alpha_activity),
            total_beta_activity=_make_var(x.beta_activity),
            total_gamma_activity=_make_var(x.gamma_activity),
            total_mass=_make_var(x.total_mass),
            total_heat=_make_var(x.total_heat),
            total_alpha_heat=_make_var(x.alpha_heat),
            total_beta_heat=_make_var(x.beta_heat),
            total_gamma_heat=_make_var(x.gamma_heat),
            total_ingest1ion_dose=_make_var(x.ingestion_dose),
            total_inhalation_dose=_make_var(x.inhalation_dose),
            total_dose_rate=_make_var(x.dose_rate.dose),
            nuclide_half_life=_make_nuclide_var(lambda n: n.half_life, x.nuclides),
            # atomic_mass=_make_nuclide_var(lambda n: n.key().atomic_mass, x.nuclides),
            nuclide_atoms=_make_time_step_and_nuclide_var(
                lambda n: n.atoms, x.nuclides
            ),
            nuclide_grams=_make_time_step_and_nuclide_var(
                lambda n: n.grams, x.nuclides
            ),
            nuclide_activity=_make_time_step_and_nuclide_var(
                lambda n: n.activity, x.nuclides
            ),
            nuclide_alpha_activity=_make_time_step_and_nuclide_var(
                lambda n: n.alpha_activity, x.nuclides
            ),
            nuclide_beta_activity=_make_time_step_and_nuclide_var(
                lambda n: n.beta_activity, x.nuclides
            ),
            nuclide_gamma_activity=_make_time_step_and_nuclide_var(
                lambda n: n.gamma_activity, x.nuclides
            ),
            nuclide_heat=_make_time_step_and_nuclide_var(lambda n: n.heat, x.nuclides),
            nuclide_alpha_heat=_make_time_step_and_nuclide_var(
                lambda n: n.alpha_heat, x.nuclides
            ),
            nuclide_beta_heat=_make_time_step_and_nuclide_var(
                lambda n: n.beta_heat, x.nuclides
            ),
            nuclide_gamma_heat=_make_time_step_and_nuclide_var(
                lambda n: n.gamma_heat, x.nuclides
            ),
            nuclide_dose=_make_time_step_and_nuclide_var(lambda n: n.dose, x.nuclides),
            nuclide_ingestion=_make_time_step_and_nuclide_var(
                lambda n: n.ingestion, x.nuclides
            ),
            nuclide_inhalation=_make_time_step_and_nuclide_var(
                lambda n: n.inhalation, x.nuclides
            ),
        )

        coords = dict(
            time_step_number=_make_var(x.number, dtype=int),
            elapsed_time=_make_var(x.elapsed_time),
            nuclide=nuclide_coordinate,
            # z=_make_nuclide_var(lambda n: n.key().z, x.nuclides, dtype=int),
            zai=_make_nuclide_var(lambda n: n.zai, x.nuclides, dtype=int),
        )

        if x.gamma_spectrum is not None:
            gamma_boundaries = x.gamma_spectrum.boundaries
            gamma_value = x.gamma_spectrum.values
            assert len(gamma_value) + 1 == len(gamma_boundaries)
            gamma_value = np.insert(gamma_value, 0, 0.0).reshape(
                1, len(gamma_boundaries)
            )
            data_vars["gamma"] = (("time_step_number", "gamma_boundaries"), gamma_value)
            coords["gamma_boundaries"] = gamma_boundaries

        tsr = xr.Dataset(data_vars=data_vars, coords=coords)
        _ds = xr.merge([_ds, tsr])
        return _ds

    ds = reduce(_add_time_step_record, inventory, xr.Dataset())

    meta_info = inventory.meta_info

    ds.coords["timestamp"] = pd.date_range(meta_info.timestamp, periods=1)
    ds.coords["timestamp"].attrs[
        "long description"
    ] = "FISPACT datasets can be merged and then selected by timestamp as coordinate"

    ds.attrs["run_name"] = meta_info.run_name
    ds.attrs["flux_name"] = meta_info.flux_name
    ds.attrs["dose_rate_type"] = meta_info.dose_rate_type
    ds.attrs["dose_rate_distance"] = meta_info.dose_rate_distance

    ds.irradiation_time.attrs["units"] = "s"
    ds.cooling_time.attrs["units"] = "s"
    ds.duration.attrs["units"] = "s"
    ds.elapsed_time.attrs["units"] = "s"
    ds.flux.attrs["units"] = "n/cm^2/s"

    if "gamma_boundaries" in ds.coords.keys():
        ds.coords["gamma_boundaries"].attrs["units"] = "MeV"

    ds.total_activity.attrs["units"] = "Bq"
    ds.total_alpha_activity.attrs["units"] = "Bq"
    ds.total_beta_activity.attrs["units"] = "Bq"
    ds.total_gamma_activity.attrs["units"] = "Bq"

    ds.total_mass.attrs["units"] = "kg"

    ds.total_heat.attrs["units"] = "kW"
    ds.total_alpha_heat.attrs["units"] = "kW"
    ds.total_beta_heat.attrs["units"] = "kW"
    ds.total_gamma_heat.attrs["units"] = "kW"

    ds.total_ingest1ion_dose.attrs["units"] = "Sv"
    ds.total_inhalation_dose.attrs["units"] = "Sv"
    ds.total_dose_rate.attrs["units"] = "Sv/hr"

    ds.nuclide_half_life.attrs["units"] = "s"
    # ds.nuclide_atomic_mass.attrs["units"] = "au"
    ds.nuclide_grams.attrs["units"] = "g"

    ds.nuclide_activity.attrs["units"] = "Bq"
    ds.nuclide_alpha_activity.attrs["units"] = "Bq"
    ds.nuclide_beta_activity.attrs["units"] = "Bq"
    ds.nuclide_gamma_activity.attrs["units"] = "Bq"

    ds.nuclide_heat.attrs["units"] = "kW"
    ds.nuclide_alpha_heat.attrs["units"] = "kW"
    ds.nuclide_beta_heat.attrs["units"] = "kW"
    ds.nuclide_gamma_heat.attrs["units"] = "kW"

    ds.nuclide_dose.attrs["units"] = "Sv"
    ds.nuclide_ingestion.attrs["units"] = "Sv"
    ds.nuclide_inhalation.attrs["units"] = "Sv"

    # Restore int type after all merges.
    # see https://xarray.pydata.org/en/stable/user-guide/combining.html
    # Note that  due  to the underlying representation of
    # missing values as floating point numbers(NaN),
    # variable  data type is not always  preserved when merging in this manner.
    ds.coords["zai"] = ("nuclide", np.asarray(ds.coords["zai"].values, dtype=int))

    return ds


def get_timestamp(ds: xr.Dataset):
    return ds.timestamp[0].dt


def get_atomic_numbers(ds: xr.Dataset) -> np.ndarray:
    def mapper(x):
        return SYMBOL_2_ATOMIC_NUMBER[x.item()[0]]

    return np.fromiter(map(mapper, ds.nuclide), dtype=int)


def add_atomic_number_coordinate(
    ds: xr.Dataset, coordinate_name="atomic_number"
) -> xr.Dataset:
    ds.coords[coordinate_name] = ("nuclide", get_atomic_numbers(ds))
    return ds


def get_atomic_masses(ds: xr.Dataset) -> np.ndarray:
    def mapper(x):
        element, mass_number, _ = x.item()
        return get_nuclide_mass(element, mass_number)

    return np.fromiter(map(mapper, ds.nuclide), dtype=int)


def add_atomic_masses(ds: xr.Dataset, variable_name="atomic_mass") -> xr.Dataset:
    ds[variable_name] = ("nuclide", get_atomic_masses(ds))
    return ds


def scale_by_flux(ds: xr.Dataset, scale: float, columns=None) -> xr.Dataset:
    """Normalize irradiation results for actual flux value.

    A FISPACT irradiation scenario may use flux value,
    which is different from actual one.
    This is the case for standard ITER SA2 scenario.
    For flux, only difference with initial activation properties is to be scaled.

    Notes:
        Assuming that the first time step presents initial values.
    """
    initial = ds.sel(time_step_number=1).fillna(0.0)

    ds = scale_by_mass(ds - initial, scale, columns) + initial

    return ds


def scale_by_mass(ds: xr.Dataset, scale: float, columns=None) -> xr.Dataset:
    """All the activation values including initial can be scaled to actual weight."""
    if columns is None:
        columns = SCALABLE_COLUMNS
    return ds[columns] * scale

    # TODO dvp: check if the dose rate is not always computed for 1g in v.5
    # "time_step_ingestion_dose",
    # "time_step_inhalation_dose",
    # "ingestion",
    # "inhalation",
    # if ds.attrs["dose_mode"] == "Point source":
    #     ds.time_step_dose *= scale
    #     ds.dose *= scale


def save_nc(ds: xr.Dataset, cn: Union[str, Path], **kwargs) -> None:
    """Save a dataset with nuclide index converted.

    See: https://github.com/pydata/xarray/issues/1077
    """
    to_save = ds.reset_index("nuclide")
    if "engine" not in kwargs:
        kwargs["engine"] = "h5netcdf"
    to_save.to_netcdf(cn, **kwargs)


def load_nc(cn: Union[str, Path], **kwargs) -> xr.Dataset:
    if "engine" not in kwargs:
        kwargs["engine"] = "h5netcdf"
    ds = xr.load_dataset(cn, **kwargs)
    ds = ds.set_index(nuclide=["element", "mass_number", "state"])
    return ds


def from_json(path: Union[str, Path, TextIO]) -> xr.Dataset:
    return create_dataset(inventory_from_json(path))



"""The Class to represent FISPACT fluxes file content."""

from typing import Callable, TextIO, Tuple

from dataclasses import dataclass
from io import StringIO, TextIOWrapper
from pathlib import Path

import numpy as np

from multipledispatch import dispatch
from numpy import allclose, array_equal
from numpy import ndarray as array
from xpypact.utils.io import print_cols

LAST_TWO_DECADES = np.fromstring(
    """
        1.0000E+9
        9.6000E+8
        9.2000E+8
        8.8000E+8
        8.4000E+8
        8.0000E+8
        7.6000E+8
        7.2000E+8
        6.8000E+8
        6.4000E+8
        6.0000E+8
        5.6000E+8
        5.2000E+8
        4.8000E+8
        4.4000E+8
        4.0000E+8
        3.6000E+8
        3.2000E+8
        2.8000E+8
        2.4000E+8
        2.0000E+8
        1.8000E+8
        1.6000E+8
        1.5000E+8
        1.4000E+8
        1.3000E+8
        1.2000E+8
        1.1000E+8
        1.0000E+8
        9.0000E+7
        8.0000E+7
        7.5000E+7
        7.0000E+7
        6.5000E+7
        6.0000E+7
        5.8000E+7
        5.6000E+7
        5.4000E+7
        5.2000E+7
        5.0000E+7
        4.8000E+7
        4.6000E+7
        4.4000E+7
        4.2000E+7
        4.0000E+7
        3.8000E+7
        3.6000E+7
        3.4000E+7
        3.2000E+7
        3.0000E+7
        2.9000E+7
        2.8000E+7
        2.7000E+7
        2.6000E+7
        2.5000E+7
        2.4000E+7
        2.3000E+7
        2.2000E+7
        2.1000E+7
        2.0000E+7
        1.9800E+7
        1.9600E+7
        1.9400E+7
        1.9200E+7
        1.9000E+7
        1.8800E+7
        1.8600E+7
        1.8400E+7
        1.8200E+7
        1.8000E+7
        1.7800E+7
        1.7600E+7
        1.7400E+7
        1.7200E+7
        1.7000E+7
        1.6800E+7
        1.6600E+7
        1.6400E+7
        1.6200E+7
        1.6000E+7
        1.5800E+7
        1.5600E+7
        1.5400E+7
        1.5200E+7
        1.5000E+7
        1.4800E+7
        1.4600E+7
        1.4400E+7
        1.4200E+7
        1.4000E+7
        1.3800E+7
        1.3600E+7
        1.3400E+7
        1.3200E+7
        1.3000E+7
        1.2800E+7
        1.2600E+7
        1.2400E+7
        1.2200E+7
        1.2000E+7
        1.1800E+7
        1.1600E+7
        1.1400E+7
        1.1200E+7
        1.1000E+7
        1.0800E+7
        1.0600E+7
        1.0400E+7
        1.0200E+7
    """,
    dtype=float,
    sep=" \n",
)[::-1]


def compute_709_bins() -> np.ndarray:
    """Computes bins for 709 groups according to the specification.

    Decades from -5 to 9 are divided to 50 bins equidistant in log scale.

    Note:
        The computed bins show better correspondence with FISPACT converted fluxes,
        then available tabulated data.
        See notebooks/dvp/demo_fispact_flux_generator.ipynb

    Returns:
        np.ndarray: array of 709 bins
    """
    template = np.logspace(0, 1, 51)
    res = np.hstack([template[1:] * 10**i for i in range(-5, 7)] + [LAST_TWO_DECADES])
    return np.insert(res, 0, 1e-5)


FISPACT_709_BINS = compute_709_bins()


@dataclass(eq=False, order=False)
class Fluxes:
    """Class to represent flux data."""

    energy_bins: array
    fluxes: array
    comment: str
    norm: float = 1.0

    def __post_init__(self) -> None:
        """Validate sizes of arrays.

        Raises:
            ValueError: if sizes of bins and fluxes are not compatible.
        """
        if self.fluxes.size != self.energy_bins.size - 1:
            raise ValueError(
                "Incompatible sizes of bins and fluxes, "
                f"{self.fluxes.size} and {self.energy_bins.size}"
            )

    @property
    def total(self) -> float:
        """Return flux total.

        Returns:
            float: flux total
        """
        return np.sum(self.fluxes).item()

    def __hash__(self) -> int:
        """Use available information for hash.

        Returns:
            hash
        """
        return hash((self.energy_bins.size, self.fluxes.size, self.comment, self.norm))

    def __eq__(self, other: "Fluxes") -> bool:
        """Compare fluxes.

        Args:
            other: other fluxes

        Returns:
            bool: True, if the fluxes are equal, otherwise - False
        """
        return self is other or (
            array_equal(self.energy_bins, other.energy_bins)
            and array_equal(self.fluxes, other.fluxes)
            and self.comment == other.comment
            and self.norm == other.norm
        )

    def __ne__(self, other: "Fluxes") -> bool:
        """Compare fluxes.

        Args:
            other: other fluxes

        Returns:
            bool: False, if the fluxes are equal, otherwise - True
        """
        return not self.__eq__(other)


def is_709_fluxes(fluxes: Fluxes) -> bool:
    """Check if fluxes are 709-kind of fluxes.

    This fluxes can be output withou energy bins and
    can be used with 'flux' statement in FISPACT run configuration.

    Args:
        fluxes: what to check

    Returns:
        bool: True, if fluxes are 709 kind of fluxes.
    """
    return fluxes.energy_bins is FISPACT_709_BINS and fluxes.fluxes.size == 709


def are_fluxes_equal(a: Fluxes, b: Fluxes) -> bool:
    """Compare data of fluxes.

    Check if fluxes are equivalent by data, disregarding differences in comment and norm.

    Args:
        a: the first flux to compare
        b: the second flux to compare

    Returns:
        bool: True, if the fluxes are equivalent by data
    """
    return array_equal(a.energy_bins, b.energy_bins) and array_equal(a.fluxes, b.fluxes)


def are_fluxes_close(
    a: Fluxes,
    b: Fluxes,
    rtol: float = 1.0e-5,
    atol: float = 1.0e-8,
    equal_nan: bool = False,
) -> bool:
    """Compare data of fluxes approximately.

    Check if fluxes are equivalent by data with some tolerance, disregarding differences in comment and norm.

    Args:
        a: the first flux to compare
        b: the second flux to compare
        rtol: relative tolerance
        atol: absolute tolerance

    Returns:
        bool: True, if the fluxes are equivalent by data
    """
    return allclose(
        a.energy_bins, b.energy_bins, rtol=rtol, atol=atol, equal_nan=equal_nan
    ) and allclose(a.fluxes, b.fluxes, rtol=rtol, atol=atol, equal_nan=equal_nan)


def read_fluxes(
    stream: TextIO, define_bins_and_fluxes: Callable[[array], Tuple[array, array]]
) -> Fluxes:
    lines = stream.readlines()
    comment = lines[-1].strip()
    norm = float(lines[-2])
    lines = lines[:-2]
    data = np.fromstring(" ".join(lines), dtype=float, count=-1, sep=" ")
    energy_bins, fluxes = define_bins_and_fluxes(data)
    return Fluxes(energy_bins, fluxes, comment, norm)


@dispatch(TextIOWrapper)
def read_arb_fluxes(
    stream: TextIO,
) -> Fluxes:
    return read_fluxes(stream, define_arb_bins_and_fluxes)


@dispatch(Path)
def read_arb_fluxes(path: Path) -> Fluxes:
    with path.open() as stream:
        return read_arb_fluxes(stream)


@dispatch(str)
def read_arb_fluxes(text: str) -> Fluxes:
    with StringIO(text) as stream:
        return read_fluxes(stream, define_arb_bins_and_fluxes)


@dispatch(TextIOWrapper)
def read_709_fluxes(
    stream: TextIO,
) -> Fluxes:
    return read_fluxes(stream, define_709_bins_and_fluxes)


@dispatch(Path)
def read_709_fluxes(path: Path) -> Fluxes:
    with path.open() as stream:
        return read_709_fluxes(stream)


@dispatch(str)
def read_709_fluxes(text: str) -> Fluxes:
    with StringIO(text) as stream:
        return read_709_fluxes(stream)


def define_arb_bins_and_fluxes(data: array) -> Tuple[array, array]:
    sz = data.size
    assert sz & 1, "The number of float values from arb_flux file should be odd"
    bins_end = sz // 2 + 1
    energy_bins = data[:bins_end]
    fluxes = data[bins_end:]
    energy_bins = energy_bins[::-1]
    fluxes = fluxes[::-1]
    return energy_bins, fluxes


def define_709_bins_and_fluxes(data: array) -> Tuple[array, array]:
    assert data.size == 709, "Invalid data for FISPACT fluxes"
    return FISPACT_709_BINS, data[::-1]


# def make_709_fluxes(fluxes: Fluxes):
#     energies, hist, comment, norm = (
#         fluxes.energy_bins,
#         fluxes.fluxes,
#         fluxes.comment,
#         fluxes.norm,
#     )
#     e_min_max = energies[:: energies.size - 1]
#     i_min, i_max = np.searchsorted(FISPACT_709_BINS, e_min_max)
#     if FISPACT_709_BINS[i_max] == e_min_max[1]:
#         i_max += 1
#     rebin_hist = ut.rebin_1d(
#         hist,
#         np.log(energies),
#         np.log(FISPACT_709_BINS[i_min:i_max]),
#         grouped=True,
#         assume_sorted=True,
#     )
#     leading_zeros = np.zeros(i_min, dtype=float)
#     trailing_zeros = np.zeros(709 - i_max + 1, dtype=float)
#     assert leading_zeros.size + rebin_hist.size + trailing_zeros.size == 709
#     rebin_hist = np.hstack(
#         [
#             leading_zeros,
#             rebin_hist,
#             trailing_zeros,
#         ]
#     )
#     result = Fluxes(FISPACT_709_BINS, rebin_hist, comment, norm)
#     assert is_709_fluxes(result)
#     return result


def print_fluxes(fluxes: Fluxes, fid: TextIO, arbitrary: bool, max_columns=7):
    if arbitrary:
        sequence = fluxes.energy_bins[::-1]
        column = print_cols(sequence, fid, max_columns, fmt="{:g}")
        if column != 0:
            print(file=fid)
    sequence = fluxes.fluxes[::-1]
    column = print_cols(sequence, fid, max_columns)
    if column != 0:
        print(file=fid)
    print(fluxes.norm, file=fid)
    print(fluxes.comment, file=fid, end="")


# def print_709_fluxes(fluxes: Fluxes, fid: TextIO, max_columns=7):
#     if not is_709_fluxes(fluxes):
#         fluxes = make_709_fluxes(fluxes)
#     print_fluxes(fluxes, fid, False, max_columns)


def print_arbitrary_fluxes(fluxes: Fluxes, fid: TextIO, max_columns=7) -> None:
    """Print fluxes in FISPACT arbitrary flux format.

    Args:
        fluxes: what to print
        fid: output stream
        max_columns: max number of columns in a row

    """
    print_fluxes(fluxes, fid, True, max_columns)



"""Nuclide specification in FISPACT JSON."""

from typing import Dict

from dataclasses import dataclass

try:
    from scipy.constants import Avogadro
except ImportError:
    Avogadro = 6.02214075999999987023872e23


@dataclass
class Nuclide:
    """Nuclide properties from FISPACT JSON."""

    element: str
    isotope: int
    state: str = ""
    zai: int = 0  # Introduced in FISPACT-II v5: z*10000 + a*10 + (state? 1: 0)
    half_life: float = 0.0
    atoms: float = 0.0
    grams: float = 0.0
    activity: float = 0.0
    alpha_activity: float = 0.0  # FISPACT 5.0
    beta_activity: float = 0.0  # -/-
    gamma_activity: float = 0.0  # -/-
    heat: float = 0.0
    alpha_heat: float = 0.0
    beta_heat: float = 0.0
    gamma_heat: float = 0.0
    dose: float = 0.0
    ingestion: float = 0.0
    inhalation: float = 0.0

    @classmethod
    def from_json(cls, json_dict: Dict) -> "Nuclide":
        return cls(**json_dict)



"""FISPACT run title data class."""
from typing import Dict

from dataclasses import dataclass


@dataclass
class RunData:
    """FISPACT run title data."""

    timestamp: str
    run_name: str
    flux_name: str

    @classmethod
    def from_json(cls, json_dict: Dict) -> "RunData":
        """Retrieve data from JSON."""
        return cls(**json_dict)



"""The `xpypact` package.

Wraps FISPACT workflow. Transforms FISPACT output to xarray datasets.
"""
from importlib import metadata as _meta
from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

__distribution__ = _meta.distribution(__name__)
__meta_data__ = __distribution__.metadata
__author__ = __meta_data__["Author"]
__author_email__ = __meta_data__["Author-email"]
__license__ = __meta_data__["License"]
__summary__ = __meta_data__["Summary"]
__copyright__ = f"Copyright 2021 {__author__}"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['xpypact', 'xpypact.utils']

package_data = \
{'': ['*']}

install_requires = \
['h5netcdf>=0.13.1,<0.14.0',
 'mckit-nuclides>=0.1.1,<0.2.0',
 'multipledispatch>=0.6.0,<0.7.0',
 'numpy>=1.21.4,<2.0.0',
 'openpyxl>=3.0.9,<4.0.0',
 'orjson>=3.6.7,<4.0.0',
 'pandas>=1.3.4,<2.0.0',
 'xarray>=0.21.1,<0.22.0']

setup_kwargs = {
    'name': 'xpypact',
    'version': '0.1.2',
    'description': '"Python tools to work with elements and isotopes"',
    'long_description': '==============================================================================\n*xpypact*: FISPACT wrapper and output to datasets converter\n==============================================================================\n\n\n\n|Maintained| |License| |Versions| |PyPI| |Docs|\n\n.. contents::\n\n\nNote:\n    This document is in progress.\n\nDescription\n-----------\n\nThe module configures and runs FISPACT, converts FISPACT output to xarray datasets.\n\n.. TODO dvp: apply FISPACT v.5 API and describe here.\n\n\nInstallation\n------------\n\n.. TODO dvp: check and report all possible ways to install (pip, poetry)\n\n\nExamples\n--------\n\n.. TODO\n\nContributing\n------------\n\n.. image:: https://github.com/MC-kit/xpypact/workflows/Tests/badge.svg\n   :target: https://github.com/MC-kit/xpypact/actions?query=workflow%3ATests\n   :alt: Tests\n.. image:: https://codecov.io/gh/MC-kit/xpypact/branch/master/graph/badge.svg?token=wlqoa368k8\n  :target: https://codecov.io/gh/MC-kit/xpypact\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n.. image:: https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336\n   :target: https://pycqa.github.io/isort/\n.. image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n\nReferences\n----------\n\n.. TODO dvp: add references to FISPACT, pypact and used libraries:  poetry, xarray etc\n\n\n.. Substitutions\n\n.. |Maintained| image:: https://img.shields.io/badge/Maintained%3F-yes-green.svg\n   :target: https://github.com/MC-kit/xpypact/graphs/commit-activity\n.. |Tests| image:: https://github.com/MC-kit/xpypact/workflows/Tests/badge.svg\n   :target: https://github.com/MC-kit/xpypact/actions?workflow=Tests\n   :alt: Tests\n.. |License| image:: https://img.shields.io/github/license/MC-kit/xpypact\n   :target: https://github.com/MC-kit/xpypact\n.. |Versions| image:: https://img.shields.io/pypi/pyversions/xpypact\n   :alt: PyPI - Python Version\n.. |PyPI| image:: https://img.shields.io/pypi/v/xpypact\n   :target: https://pypi.org/project/xpypact/\n   :alt: PyPI\n.. |Docs| image:: https://readthedocs.org/projects/xpypact/badge/?version=latest\n   :target: https://xpypact.readthedocs.io/en/latest/?badge=latest\n   :alt: Documentation Status\n',
    'author': 'dvp',
    'author_email': 'dmitri_portnov@yahoo.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MC-kit/xpypact',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
